/*
 * Embedded Linux library
 * Copyright (C) 2012  Intel Corporation
 * Copyright (C) 2023  Cruise LLC
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/**
 * SECTION:minheap
 * @short_description: Minimum Heap Support
 *
 * Minimum Heap support
 */
